@@ -17,10 +17,11 @@
 #ifndef GLOBUS_I_GRIDFTP_SERVER_H
 #define GLOBUS_I_GRIDFTP_SERVER_H
 
-#include "config.h"
+#include "globus_i_gridftp_server_config.h"
 #include "globus_gridftp_server.h"
 #include "globus_gridftp_server_control.h"
 #include "globus_i_gfs_acl.h"
+#include "globus_i_gfs_http.h"
 #include "globus_xio.h"
 #include "globus_xio_system.h"
 #include "globus_xio_tcp_driver.h"
@@ -29,6 +30,16 @@
 #include "globus_gsi_authz.h"
 #include "globus_usage.h"
 
+#define _GSSL(s) globus_common_i18n_get_string_by_key(\
+		    NULL, \
+		    "globus_gridftp_server", \
+		    s)
+
+#define _FSSL(s,p) globus_common_i18n_get_string_by_key(\
+		     p, \
+		    "globus_gridftp_server", \
+		    s)
+
 typedef void
 (*globus_i_gfs_server_close_cb_t)(
     void *                              user_arg,
