@@ -14,13 +14,55 @@
  * limitations under the License.
  */
 
+#include "globus_common.h"
 #include "globus_gridftp_server.h"
 #include "globus_xio.h"
 #include "globus_xio_file_driver.h"
-#include "openssl/md5.h"
+#include <openssl/md5.h>
 #include "version.h"
-#include <grp.h>
+
 #include <utime.h>
+#ifndef TARGET_ARCH_WIN32
+#include <grp.h>
+#endif
+
+#ifdef TARGET_ARCH_WIN32
+#define S_ISLNK(x) 0
+#define lstat(x,y) stat(x,y)
+#define mkdir(x,y) mkdir(x)
+#define chown(x,y,z) -1
+#define symlink(x,y) -1
+#define readlink(x,y,z) 0
+#define realpath(x,y) strcpy(y,x)
+#define scandir(a,b,c,d) 0
+#define alphasort(x,y) 0
+#endif
+
+#ifdef TARGET_ARCH_WIN32
+#define getuid() 1
+#define getpwuid(x) 0
+#define initgroups(x,y) -1
+#define getgroups(x,y) -1
+#define setgroups(x,y) 0
+#define setgid(x) 0
+#define setuid(x) 0
+#define sync() 0
+#define fork() -1
+#define setsid() -1
+#define chroot(x) -1
+#define globus_libc_getpwnam_r(a,b,c,d,e) -1
+#define globus_libc_getpwuid_r(a,b,c,d,e) -1
+#endif
+
+#ifdef TARGET_ARCH_WIN32
+
+#define getpwnam(x) 0
+
+#define getgrgid(x) 0
+#define getgrnam(x) 0
+
+#endif
+
 
 #ifndef MAXPATHLEN
 #define MAXPATHLEN 4096
@@ -565,11 +607,17 @@ globus_l_gfs_file_partition_path(
     buf[MAXPATHLEN - 1] = '\0';
     
     filepart = strrchr(buf, '/');
+    
+#ifdef WIN32
+
+
+#else
     while(filepart && !*(filepart + 1) && filepart != buf)
     {
         *filepart = '\0';
         filepart = strrchr(buf, '/');
     }
+#endif
 
     if(!filepart)
     {
@@ -602,6 +650,11 @@ globus_l_gfs_file_partition_path(
         }
     }    
 
+    if(filename[0] == 0)
+    { 
+        filename[0] = '/';
+        filename[1] = 0;
+    }
     GlobusGFSFileDebugExit();
 }
 
@@ -784,6 +837,228 @@ globus_l_gfs_file_stat(
         stat_count = 1;
     }
     else
+#ifdef WIN32
+    {
+        struct dirent *                 dir_entry;
+        int                             i;
+        char                            dir_path[MAXPATHLEN];
+        int                             stat_limit_check = GFS_STAT_COUNT_CHECK;
+        int                             stat_limit_max = GFS_STAT_COUNT_MAX;
+        time_t                          stat_limit_time;
+        globus_bool_t                   check_cdir = GLOBUS_TRUE;
+
+        stat_limit_time = time(NULL) + GFS_STAT_TIME;
+    
+        stat_count = stat_info->include_path_stat ? 1 : 0;
+
+    
+        {
+            if(stat_info->pathname && 
+            stat_info->pathname[0] == '/' && stat_info->pathname[1] == 0)
+            {
+                DWORD drivemask;
+                char drive[] = "A";
+                
+                stat_array = (globus_gfs_stat_t *)
+                    globus_malloc(sizeof(globus_gfs_stat_t) * 30);
+                if(!stat_array)
+                {
+                    result = GlobusGFSErrorMemory("stat_array");
+                    goto error_alloc2;
+                }
+
+                drivemask = GetLogicalDrives();
+                while(drivemask && *drive <= 'Z')
+                {
+                    if(drivemask & 1)
+                    {
+                        stat_buf.st_dev = *drive;
+                        globus_l_gfs_file_copy_stat(&stat_array[stat_count], &stat_buf, drive, NULL, 0, 0);
+                        stat_count++;
+                    }
+                    (*drive)++;
+                    drivemask >>= 1;
+                }
+
+                goto done_fake;
+                
+            }
+        }
+
+
+        dir = globus_libc_opendir(stat_info->pathname);
+        if(!dir)
+        {
+            result = GlobusGFSErrorSystemError("opendir", errno);
+            if (!stat_info->include_path_stat)
+            goto error_open;
+        }
+        
+        total_stat_count = 0;
+
+        while(globus_libc_readdir_r(dir, &dir_entry) == 0 && dir_entry)
+        {
+            total_stat_count++;
+            globus_free(dir_entry);
+        }
+        
+        globus_libc_rewinddir(dir);
+        
+        stat_array = (globus_gfs_stat_t *)
+            globus_malloc(sizeof(globus_gfs_stat_t) * GLOBUS_MIN((stat_limit_max+1), total_stat_count));
+        if(!stat_array)
+        {
+            result = GlobusGFSErrorMemory("stat_array");
+            goto error_alloc2;
+        }
+        
+        snprintf(
+            dir_path, 
+            sizeof(dir_path), 
+            "%s/%s", 
+            (basepath[0] != '/' || basepath[1] != '\0') ? basepath : "", 
+            filename);
+            
+        dir_path[MAXPATHLEN - 1] = '\0';
+        if(!basepath[0] && filename[0] == '/')
+        {
+            dir_path[0] = 0;
+        }
+        
+        i = 0;
+        if(stat_info->include_path_stat) 
+        {
+            globus_l_gfs_file_copy_stat(&stat_array[i++], &stat_buf, filename, NULL, 0,
+                dir ? GLOBUS_GRIDFTP_SERVER_CONTROL_STAT_SUCCESS : GLOBUS_GRIDFTP_SERVER_CONTROL_STAT_OPENFAILED);
+        }
+        
+        while(globus_libc_readdir_r(dir, &dir_entry) == 0 && dir_entry)
+        {
+            char                        path[MAXPATHLEN];
+                
+                base_error = GLOBUS_GRIDFTP_SERVER_CONTROL_STAT_SUCCESS;
+            snprintf(path, sizeof(path), "%s/%s", dir_path, dir_entry->d_name);
+            path[MAXPATHLEN - 1] = '\0';
+        
+            /* lstat is the same as stat when not operating on a link */
+            if(lstat(path, &stat_buf) != 0)
+            {
+                globus_free(dir_entry);
+                /* just skip invalid entries */
+                total_stat_count--;
+                continue;
+            }
+            /* if this is a link we still need to stat to get the info we are 
+                interested in and then use realpath() to get the full path of 
+                the symlink target */
+            *symlink_target = '\0';
+            if(S_ISLNK(stat_buf.st_mode))
+            {
+                int stat_result = 0;
+            
+                if(stat_info->use_symlink_info)
+                {
+                    memset(&link_stat_buf, 0, sizeof(struct stat));
+                    stat_result = stat(path, &link_stat_buf);
+                }
+                else if(stat(path, &stat_buf) != 0)
+                {
+                    globus_free(dir_entry);
+                    /* just skip invalid entries */
+                    total_stat_count--;
+                    continue;
+                }
+                if(stat_result < 0 || realpath(path, symlink_target) == NULL)
+                {
+                    int nchars = readlink(path, symlink_target, MAXPATHLEN);
+                    if (nchars < 0)
+                    {
+                        globus_free(dir_entry);
+                        /* just skip invalid entries */
+                        total_stat_count--;
+                        continue;
+                    }
+                    symlink_target[nchars] = '\0';
+                    base_error = GLOBUS_GRIDFTP_SERVER_CONTROL_STAT_INVALIDLINK;
+                }
+            }
+            
+            globus_l_gfs_file_copy_stat(
+                    &stat_array[i], &stat_buf, dir_entry->d_name, symlink_target, link_stat_buf.st_mode, base_error);
+
+            /* set nlink to total files in dir for . entry */
+            if(check_cdir && dir_entry->d_name && 
+                dir_entry->d_name[0] == '.' && dir_entry->d_name[1] == '\0')
+            {
+                check_cdir = GLOBUS_FALSE;
+                stat_array[i].nlink = total_stat_count;
+            }
+            if(stat_array[i].ino == 0)
+            {
+                unsigned long                       h = 0;
+                char *                              key;
+            
+                key = path;
+                while(*key)
+                {
+                    h = 131 * h + *key++;
+                }
+                stat_array[i].ino = h;
+            }
+
+            i++;
+            globus_free(dir_entry);
+            
+            /* send updates every GFS_STAT_TIME, checked every GFS_STAT_CHECK */
+            if(i >= stat_limit_check)
+            {
+                time_t                  tmp_time;
+                globus_bool_t           send_stats = GLOBUS_FALSE;
+                
+                tmp_time = time(NULL);
+                if(i >= stat_limit_max || tmp_time > stat_limit_time)
+                {
+                    send_stats = GLOBUS_TRUE;
+                }
+                else
+                {
+                    stat_limit_check += GFS_STAT_COUNT_CHECK;
+                }
+                
+                if(send_stats)
+                {
+                    stat_count = i;
+                    stat_limit_check = GFS_STAT_COUNT_CHECK;
+                    stat_limit_time = tmp_time + GFS_STAT_TIME;
+
+                    i = 0;
+                    
+                    globus_gridftp_server_finished_stat_partial(
+                        op, GLOBUS_SUCCESS, stat_array, stat_count);
+                        
+                    globus_l_gfs_file_destroy_stat(stat_array, stat_count);
+                    
+                    stat_array = (globus_gfs_stat_t *) globus_malloc(
+                        sizeof(globus_gfs_stat_t) * (stat_limit_max + 1));
+                    if(!stat_array)
+                    {
+                        result = GlobusGFSErrorMemory("stat_array");
+                        goto error_alloc2;
+                    }
+                    
+                    stat_count = 0;
+                }
+            }                
+        }
+        stat_count = i;
+        
+        closedir(dir);
+        
+
+    }
+done_fake:
+
+#else
     {
         struct dirent *                 dir_entry;
         struct dirent **                entries = NULL;
@@ -887,7 +1162,7 @@ globus_l_gfs_file_stat(
                 check_cdir = GLOBUS_FALSE;
                 stat_array[i].nlink = total_stat_count;
             }
-            
+
             i++;
             globus_free(dir_entry);
             
@@ -939,7 +1214,8 @@ globus_l_gfs_file_stat(
             globus_free(entries);
         }
     }
-    
+#endif
+
     globus_gridftp_server_finished_stat(
         op, result, stat_array, stat_count);
     globus_l_gfs_file_destroy_stat(stat_array, stat_count);
@@ -958,6 +1234,72 @@ error_stat1:
     GlobusGFSFileDebugExitWithError();
 }
 
+
+static
+globus_result_t
+globus_l_gfs_file_truncate(
+    globus_gfs_operation_t              op,
+    const char *                        pathname,
+    globus_off_t                        offset)
+{
+    int                                 rc;
+    globus_result_t                     result;
+    globus_xio_system_file_t            fd;
+    struct stat                         sbuf;
+    
+    GlobusGFSName(globus_l_gfs_file_truncate);
+    GlobusGFSFileDebugEnter();
+
+    result = globus_xio_system_file_open(
+        &fd, pathname, 
+        GLOBUS_XIO_FILE_RDWR | GLOBUS_XIO_FILE_BINARY, 0);
+    if(result != GLOBUS_SUCCESS)
+    {
+        goto error;
+    }
+
+#ifdef WIN32
+    rc = stat(pathname, &sbuf);    
+#else
+    rc = fstat(fd, &sbuf);
+#endif    
+    if(rc != 0)
+    {
+        result = GlobusGFSErrorSystemError("stat", errno);
+        goto error_close;
+    }
+    
+    if(offset > sbuf.st_size)
+    {
+        result = GlobusGFSErrorGeneric(
+            "Current size is smaller than truncate size.");
+        goto error_close;
+    }
+
+    result = globus_xio_system_file_truncate(fd, offset);
+    if(result != GLOBUS_SUCCESS)
+    {
+        goto error_close;
+    }
+    
+    result = globus_xio_system_file_close(fd);
+    if(result != GLOBUS_SUCCESS)
+    {
+        goto error;
+    }
+        
+    globus_gridftp_server_finished_command(op, GLOBUS_SUCCESS, NULL);
+        
+    GlobusGFSFileDebugExit();
+    return GLOBUS_SUCCESS;
+
+error_close:
+    globus_xio_system_file_close(fd);
+error:
+    GlobusGFSFileDebugExitWithError();
+    return result;
+}
+
 static
 globus_result_t
 globus_l_gfs_file_mkdir(
@@ -1426,7 +1768,7 @@ globus_l_gfs_file_cksm_read_cb(
                 NULL,
                 NULL,
                 NULL);
-            monitor->marker_handle = NULL;
+            monitor->marker_handle = GLOBUS_NULL_HANDLE;
         }
         
         MD5_Final(md, &monitor->mdctx);
@@ -1763,6 +2105,10 @@ globus_l_gfs_file_command(
         result = globus_l_gfs_file_delete(
             op, cmd_info->pathname, GLOBUS_FALSE);
         break;
+      case GLOBUS_GFS_CMD_TRNC:
+        result = globus_l_gfs_file_truncate(
+            op, cmd_info->pathname, cmd_info->cksm_offset);
+        break;
       case GLOBUS_GFS_CMD_SITE_RDEL:
         result = globus_l_gfs_file_delete(
             op, cmd_info->pathname, GLOBUS_TRUE);
@@ -2124,14 +2470,6 @@ globus_l_gfs_file_server_read_cb(
             globus_l_gfs_file_update_concurrency(monitor);
         }        
         
-        result = globus_l_gfs_file_dispatch_write(monitor);
-        if(result != GLOBUS_SUCCESS)
-        {
-            monitor->error = GlobusGFSErrorObjWrapFailed(
-                "globus_l_gfs_file_dispatch_write", result);
-            goto error_dispatch;
-        }
-
         rc = globus_priority_q_enqueue(
             &monitor->queue, buf_info, buf_info);
         if(rc != GLOBUS_SUCCESS)
@@ -2140,6 +2478,15 @@ globus_l_gfs_file_server_read_cb(
                 "globus_priority_q_enqueue failed");
             goto error_enqueue;
         }
+
+        result = globus_l_gfs_file_dispatch_write(monitor);
+        if(result != GLOBUS_SUCCESS)
+        {
+            monitor->error = GlobusGFSErrorObjWrapFailed(
+                "globus_l_gfs_file_dispatch_write", result);
+            goto error_dispatch;
+        }
+
     }
     globus_mutex_unlock(&monitor->lock);
     
